# Copyright (c) 2006, Stephen Fisher, Susan Davidson, and Junhyong Kim, 
# University of Pennsylvania.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA.
#
# exBatchScript-3.py - sample batch script

# This script will run the query "L10.P03.R05" from the SDSC database,
# creating a 10 leaf sub-tree, and saving the output as a NEXUS
# file. The NEXUS file will contain a Paup block instructing paup to
# use neighbor joining to reconstruct the tree. The query is then run
# a second time, including the same leaves and containing a Paup block
# using parsimony for the reconstruction. The true sub-tree topology
# will not be included in the NEXUS files but will be output
# separately as a newick string, for comparison to the reconstructed
# trees.

# load standard definitions
from startup import *

# Open a read-only connection to the SDSC Oracle server, via web proxy
connectCIPRES()

# This query creates a 10 leaf tree, sampling the 1M leaf tree
# ("TREE-0"), using the data partition 3 and "6695834184220664292L" as
# the seed for the random number generator.
queryID = "L10.p03.r05"

# Load the query java object for query "L10.P03.R05."
query = getQuery(queryID)

# Add a note to the query that contains a NEXUS formatted paup
# block. This will run neighbor joining. 
query.setNotes("\n\nBEGIN PAUP;\n  nj brlens=yes breakties=random showtree=no treefile=" + queryID + ".nj.out replace=yes;\n  quit warntsave=no;\nEND;")

# Run query, creating a paup file called "L10.p03.r05.nj.in" containing
# the sequence data for the sampled tree and the paup block
# above. This query creates a 10 leaf tree, sampling the 1M leaf tree
# ("TREE-0"), using "6695834184220664292L" as the seed for the random
# number generator. Upon completion, the variable "newTree" will
# contain the 10 leaf tree Java object. A second file called
# "L10.p03.r05.nexus.newick" will be created and will contain the
# unrooted newick string representing the 10 leaf tree.
newTree = runNexusQuery(queryID, queryID + ".nj.in", 0, 0, 4, 1)

# Store start time for paup run.
startTime = time(1)

# Run paup file just generated, saving log info to
# "L10.p03.r05.nj.log". Paup output will also be displayed to screen.
execute("paup -n " + queryID + ".nj.in -l " + queryID + ".nj.log")

# Print neighbor joining runtime in seconds. 
njTime = int((time(1) - startTime) / 1000)

# Run same query but including parsimony in paup block. Intermediate
# trees will be saved in "L10.p03.r05.trees" and the final tree will
# be stored in "L10.p03.r05.pars.out"
query.setNotes("\n\nBEGIN PAUP;\n  set maxtrees=200 increase=no;\n  hsearch timelimit=432000;\n  savetrees file=" + queryID + ".trees format=altnex replace=yes;\n  contree all /strict=no majrule=yes treefile=" + queryID + ".pars.out showtree=no replace=yes;\n  quit warntsave=no;\nEND;")

# Run query, creating a paup file called "L10.p03.r05.pars.in". No
# need to create the newick file again, since we just did that above.
newTree = runNexusQuery(queryID, queryID + ".pars.in", 0, 0, 0, 1)

# Store start time for paup run.
startTime = time(1)

# Run paup file just generated, saving log info to
# "L10.p03.r05.pars.log".
execute("paup -n " + queryID + ".pars.in -l " + queryID + ".pars.log")

# Print parsimony runtime in seconds. 
print "\nParsimony run time (sec): " + str(int((time(1) - startTime) / 1000))

# Parsimony score
tmp = CrimsonMain.getExec("grep best " + queryID + ".pars.log")
print "Parsimony score: " + tmp.split()[6]

# Print neighbor joining runtime in seconds. 
print "Neighbor Joining run time (sec): " + str(njTime)

